from __future__ import annotations

from typing import Any, Literal

import unittest.mock
from pathlib import Path

import pytest
from sphinx.errors import ExtensionError

import rocm_docs.projects

from .log_fixtures import ExpectLogFixture
from .sphinx_fixtures import SITES_BASEFOLDER

TEMPLATE_FOLDER = SITES_BASEFOLDER / "templates"


def str_or_list_to_id(val: str | list[str]) -> str:
    if isinstance(val, str):
        return val
    return ",".join(val)


def create_app(
    srcdir: Path, external_projects: Any
) -> unittest.mock.NonCallableMock:
    app = unittest.mock.NonCallableMock()
    app.srcdir = srcdir
    app.config = unittest.mock.NonCallableMock()
    app.config.overrides = []
    app.config._raw_config = {
        "external_projects_current_project": "a",
        "external_projects": external_projects,
        "external_toc_template_path": TEMPLATE_FOLDER
        / "minimal"
        / ".sphinx"
        / "_toc.yml.in",
        "external_toc_path": "_toc.yml",
        "intersphinx_mapping": {},
    }
    app.config.configure_mock(**app.config._raw_config)
    return app


@pytest.mark.parametrize(
    "external_projects",
    [[], ["a"], ["b"], ["a", "b"], "all"],
    ids=str_or_list_to_id,
)
@pytest.mark.usefixtures("_no_unexpected_warnings")
def test_external_projects(
    external_projects: list[str] | Literal["all"],
    mocked_projects: dict[str, rocm_docs.projects._Project],
    tmp_path: Path,
    with_no_git_repo: ExpectLogFixture.Validator,
) -> None:
    with_no_git_repo.required = False
    app = create_app(tmp_path, external_projects)
    rocm_docs.projects._update_config(app, app.config)

    keys = (
        external_projects
        if external_projects != "all"
        else mocked_projects.keys()
    )

    expected_mapping = {
        k: (v.target, tuple(v.inventory))
        for k, v in mocked_projects.items()
        if k in keys
    }
    assert app.config.intersphinx_mapping == expected_mapping

    # Every project is available in the HTML templates and TOC, regardless of
    # the value of "external_projects"
    expected_context = {k: v.target for k, v in mocked_projects.items()}
    assert app.config.projects_context["projects"] == expected_context


@pytest.mark.usefixtures("mocked_projects", "_no_unexpected_warnings")
def test_external_projects_invalid_value(
    expect_log: ExpectLogFixture,
    with_no_git_repo: ExpectLogFixture.Validator,
    tmp_path: Path,
) -> None:
    with_no_git_repo.required = False
    app = create_app(tmp_path, "invalid_value")

    with expect_log(
        "sphinx.rocm_docs.projects",
        "ERROR",
        'Unexpected value "invalid_value" in external_projects.\n'
        'Must be set to a list of project names or "all" to enable all projects'
        " defined in projects.yaml",
    ):
        rocm_docs.projects._update_config(app, app.config)


@pytest.mark.usefixtures("_no_unexpected_warnings")
def test_external_projects_unknown_project(
    expect_log: ExpectLogFixture,
    mocked_projects: dict[str, rocm_docs.projects._Project],
    with_no_git_repo: ExpectLogFixture.Validator,
    tmp_path: Path,
) -> None:
    with_no_git_repo.required = False
    # First keys of mocked_projects
    a_defined_project = next(iter(mocked_projects.keys()))
    app = create_app(tmp_path, [a_defined_project, "unknown_project", "foo"])

    with expect_log(
        "sphinx.rocm_docs.projects",
        "ERROR",
        'Unknown projects: ["unknown_project", "foo"] in external_projects.\n'
        "Valid projects are: [{}]".format(
            ", ".join([f'"{p}"' for p in mocked_projects])
        ),
    ):
        rocm_docs.projects._update_config(app, app.config)


@pytest.mark.usefixtures("_no_unexpected_warnings")
@pytest.mark.parametrize(
    "doxygen", ["doxygen/html", {"html": "doxygen/html"}], ids=["str", "dict"]
)
def test_doxygen_html_types(
    doxygen: str | dict[str, str],
) -> None:
    result = rocm_docs.projects._Project._get_doxygen_html({"doxygen": doxygen})  # type: ignore
    assert result == "doxygen/html"


@pytest.mark.usefixtures("_no_unexpected_warnings")
@pytest.mark.parametrize(
    "doxygen_html",
    ["https:", "//has_netloc", "has#fragment", "has?query", "/absolute"],
)
def test_doxygen_html_invalid(
    doxygen_html: str,
) -> None:
    with pytest.raises(ExtensionError):
        rocm_docs.projects._Project._get_doxygen_html({"doxygen": doxygen_html})


@pytest.mark.usefixtures("_no_unexpected_warnings")
@pytest.mark.parametrize(
    "current_project",
    [None, rocm_docs.projects._Project("", [], "", [], None)],
    ids=["no_current_project", "not_set"],
)
def test_set_doxygen_html_not_defined(
    current_project: rocm_docs.projects._Project | None,
) -> None:
    app = unittest.mock.NonCallableMock()
    app.config = unittest.mock.NonCallableMock()
    app.config.doxygen_html = "must-not-be-changed"
    rocm_docs.projects._set_doxygen_html(app, current_project)
    assert app.config.doxygen_html == "must-not-be-changed"


@pytest.mark.usefixtures("_no_unexpected_warnings")
def test_set_doxygen_html_mismatched(expect_log: ExpectLogFixture) -> None:
    app = create_app(Path(""), [])
    app.config.doxygen_html = "must-not-be-changed"
    app.config._raw_config["doxygen_html"] = "must-not-be-changed"
    with expect_log(
        "sphinx.rocm_docs.projects",
        "WARNING",
        'The setting doxygen_html="must-not-be-changed"'
        ' differs from projects.yaml value: "does-not-match"',
    ):
        rocm_docs.projects._set_doxygen_html(
            app, rocm_docs.projects._Project("", [], "", [], "does-not-match")
        )
    assert app.config.doxygen_html == "must-not-be-changed"



"""Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html

For a list of options specific to rocm-docs-core, see the user guide:
https://rocm.docs.amd.com/projects/rocm-docs-core/en/latest/
"""

# Disable fetching projects.yaml, it would be the same as the local one anyway
# except if a PR modifies it. We want to test with its version in that case
external_projects_remote_repository = ""
external_projects = ["hipify", "python", "rocm-docs-core", "rocm"]
external_projects_current_project = "rocm-docs-core"

setting_all_article_info = True
all_article_info_os = []
all_article_info_author = ""
# specific settings override any general settings (eg: all_article_info_<field>)
article_pages = [
    {
        "file": "index",
        "os": ["linux", "windows"],
        "author": "Author: AMD",
        "date": "2024-07-03",
        "read-time": "2 min read",
    },
    {
        "file": "user_guide/article_info",
        "os": [],
        "author": "",
        "date": "",
        "read-time": "",
    },
    {
        "file": "developer_guide/commitizen",
    },
]

html_theme = "rocm_docs_theme"
html_theme_options = {"flavor": "rocm"}

external_toc_path = "./sphinx/_toc.yml"

extensions = ["rocm_docs", "rocm_docs.doxygen"]
doxygen_root = "demo/doxygen"
doxysphinx_enabled = True
doxygen_project = {
    "name": "doxygen",
    "path": "demo/doxygen/xml",
}

version = "1.34.0"
release = "1.34.0"
html_title = f"ROCm Docs Core {version}"
project = "ROCm Docs Core"
author = "Advanced Micro Devices, Inc."
copyright = (
    "Copyright (c) 2026 Advanced Micro Devices, Inc. All rights reserved."
)



from rocm_docs import ROCmDocs

docs_core = ROCmDocs("ROCm Docs Core Test Project - Minimal Legacy")
docs_core.setup()

for sphinx_var in ROCmDocs.SPHINX_VARS:
    globals()[sphinx_var] = getattr(docs_core, sphinx_var)

external_projects_current_project = "a"



"""Utilities for rocm-docs-core."""

from __future__ import annotations

from typing import Any

import enum
import functools
import os
import re
from pathlib import Path

import github
from git.exc import InvalidGitRepositoryError
from git.repo import Repo
from github.GithubException import UnknownObjectException


class VersionType(enum.Enum):
    """Describes how recent a version is (i.e. latest rc, or an older release)"""

    DEVELOPMENT = enum.auto()
    ROCM_LATEST_RELEASE = enum.auto()
    OTHER_LATEST_RELEASE = enum.auto()
    OLD_RELEASE = enum.auto()
    RELEASE_CANDIDATE = enum.auto()
    PREVIEW = enum.auto()


def get_path_to_docs(
    conf_path: str | os.PathLike[Any] | None = None,
    repo_path: str | os.PathLike[Any] | None = None,
) -> str:
    """Get the relative path from the repo root to the docs."""
    if conf_path is None:
        conf_path = Path()
    elif not isinstance(conf_path, Path):
        conf_path = Path(conf_path)
    if repo_path is None:
        repo_path = conf_path
    elif not isinstance(repo_path, Path):
        repo_path = Path(repo_path)
    repo = Repo(path=repo_path, search_parent_directories=True)
    return os.path.relpath(str(conf_path), repo.working_dir)


@functools.lru_cache
def get_branch(
    repo_path: str | os.PathLike[Any] | None = None,
) -> tuple[str, str]:
    """Get the branch whose tip is checked out, even if detached.

    May be overridden with the environment variable `ROCM_DOCS_REMOTE_DETAILS`.
    """
    if "ROCM_DOCS_REMOTE_DETAILS" in os.environ:
        remote_details = os.environ["ROCM_DOCS_REMOTE_DETAILS"].split(",")
        return (remote_details[0], remote_details[1])

    def get_repo_url(remote_url: str) -> str:
        ssh_pattern = re.compile(r"git@(\w+(?:\.\w+)+):(.*?)(?:\.git)?$")
        custom_ssh_pattern = re.compile(r"git@([^:]+):(.*?)(?:\.git)?$")
        http_pattern = re.compile(r"(https?://.+)\.git")
        remote_url, ssh_num_subs = ssh_pattern.subn(
            r"http://\1/\2", remote_url, count=1
        )
        if ssh_num_subs > 0:
            return remote_url
        # If ssh prefix does not follow git@xxx.xxx format,
        # the user could be using a secondary SSH key for EMU
        remote_url, custom_ssh_num_subs = custom_ssh_pattern.subn(
            r"http://github.com/\2", remote_url, count=1
        )
        if custom_ssh_num_subs > 0:
            return remote_url
        return http_pattern.sub(r"\1", remote_url, count=1)

    if os.environ.get("READTHEDOCS", ""):
        remote_url = os.environ.get("READTHEDOCS_GIT_CLONE_URL", "")
        url = get_repo_url(remote_url)
        build_type = os.environ["READTHEDOCS_VERSION_TYPE"]
        match = re.match(r"(?:.*://)?.*\.(com|io)[/:](.*)\.git", remote_url)
        assert match is not None
        repo_fqn: str = match[1]
        if build_type in ("branch", "tag"):
            return url, os.environ["READTHEDOCS_VERSION"]
        if build_type == "external":
            gh_inst = github.Github(os.environ.get("TOKEN", None))
            print("Repository URL: " + repo_fqn)
            try:
                pull = gh_inst.get_repo(repo_fqn).get_pull(
                    int(os.environ["READTHEDOCS_VERSION"])
                )
                return url, pull.head.ref
            except UnknownObjectException as err:
                if err.data["message"] == "Not Found":
                    # Possibly a private repository that we're not
                    # authenticated for, fallback
                    return (
                        url,
                        "external-" + os.environ["READTHEDOCS_VERSION"],
                    )
        # if build type is unknown try the usual strategy

    if repo_path is None:
        repo_path = Path()
    elif not isinstance(repo_path, Path):
        repo_path = Path(repo_path)
    repo = Repo(repo_path, search_parent_directories=True)
    assert not repo.bare
    for head in repo.heads:
        if head.commit == repo.head.commit:
            tracking = head.tracking_branch()
            if tracking is not None:
                remote_url = repo.remotes[tracking.remote_name].url
                remote_url = get_repo_url(remote_url)
                return remote_url, tracking.remote_head
    for remote in repo.remotes:
        for ref in remote.refs:
            if ref.commit == repo.head.commit:
                remote_url = get_repo_url(remote.url)
                return remote_url, ref.remote_head

    # Fall-back to the current branch or a fallback value if HEAD is detached
    # In this case the repository URL cannot be provided
    try:
        branch = repo.active_branch.name
    except TypeError:
        branch = "branch-not-found"

    return "", branch


__all__ = [
    "InvalidGitRepositoryError",
    "VersionType",
    "get_branch",
    "get_path_to_docs",
]



from __future__ import annotations

import logging

import pytest

from .log_fixtures import ExpectLogFixture

logger = logging.getLogger(__name__)


def test_log_stack(expect_log: ExpectLogFixture) -> None:
    with expect_log(logger.name, "WARNING", "Foo"):
        logger.warning("Foo")
        with expect_log(logger.name, "WARNING", "Bar"):
            logger.warning("Bar")
            with expect_log(logger.name, "ERROR", "Baz"):
                logger.error("Baz")
        with expect_log(logger.name, "ERROR", "Qux"):
            logger.error("Qux")


def test_scope_of_log_stack(expect_log: ExpectLogFixture) -> None:
    logger.info("Foo")
    with expect_log(logger.name, "INFO", "Foo", required=False) as res:
        pass
    logger.info("Foo")
    assert not res.found


@pytest.mark.usefixtures("_no_unexpected_warnings")
@pytest.mark.meta_invert_fixture
def test_unexpected_warnings() -> None:
    logger.warning("An unexpected warning")


@pytest.mark.usefixtures("_no_unexpected_warnings")
def test_expected_warning(
    caplog: pytest.LogCaptureFixture, expected_logs: set[logging.LogRecord]
) -> None:
    logger.warning("An expected warning")
    records = caplog.get_records("call")
    assert len(records) == 1
    expected_logs.add(records[0])



"""Core rocm_docs extension.

It enables a core set of sphinx extensions and provides good defaults for
settings. The environment provided is meant as consistent common base for
ROCm documentation projects.
"""

from __future__ import annotations

from typing import Any, Generic, TypeVar

import inspect
import os
import urllib.parse
from abc import ABC, abstractmethod

from pydata_sphinx_theme.utils import (  # type: ignore[import-untyped]
    config_provided_by_user,
)
from sphinx.application import Sphinx
from sphinx.config import Config

from rocm_docs import article_info

T = TypeVar("T")


class _ConfigUpdater(Generic[T], ABC):
    def __init__(self, default: T) -> None:
        super().__init__()
        self.default = default

    @abstractmethod
    def __call__(self, key: str, app: Sphinx) -> None:
        pass


class _ConfigExtend(_ConfigUpdater[list[T]]):
    def __call__(self, key: str, app: Sphinx) -> None:
        getattr(app.config, key).extend(self.default)


class _ConfigDefault(_ConfigUpdater[T]):
    def __call__(self, key: str, app: Sphinx) -> None:
        if not config_provided_by_user(app, key):
            setattr(app.config, key, self.default)


class _ConfigUnion(_ConfigUpdater[set[T]]):
    def __call__(self, key: str, app: Sphinx) -> None:
        getattr(app.config, key).update(self.default)


class _ConfigMerge(_ConfigUpdater[dict[str, Any]]):
    def __call__(self, key: str, app: Sphinx) -> None:
        current_setting: dict[str, Any] = getattr(app.config, key)
        for item in self.default.items():
            current_setting.setdefault(item[0], item[1])


class _DefaultSettings:
    author = _ConfigDefault(
        'Advanced Micro Devices <a href="https://">Disclaimer and'
        " Licensing Info</a>"
    )
    # pylint: disable=redefined-builtin
    copyright = _ConfigDefault("2022-2023, Advanced Micro Devices Ltd")
    # pylint: enable=redefined-builtin
    myst_enable_extensions = _ConfigUnion(
        {
            "colon_fence",
            "dollarmath",
            "fieldlist",
            "html_image",
            "replacements",
            "substitution",
        }
    )
    myst_heading_anchors = _ConfigDefault(3)
    external_toc_exclude_missing = _ConfigDefault(False)
    epub_show_urls = _ConfigDefault("footnote")
    exclude_patterns = _ConfigExtend(["_build", "Thumbs.db", ".DS_Store"])
    numfig = _ConfigDefault(False)
    linkcheck_timeout = _ConfigDefault(10)
    linkcheck_request_headers = _ConfigMerge(
        {
            r"https://docs.github.com/": {
                "User-Agent": (
                    "Mozilla/5.0 (X11; Ubuntu; Linux i686; rv:112.0)          "
                    "       Gecko/20100101 Firefox/112.0"
                )
            }
        }
    )

    @classmethod
    def update_config(cls, app: Sphinx, _: Config) -> None:
        """Update the Sphinx configuration from the default settings."""
        for name, attr in inspect.getmembers(cls):
            if isinstance(attr, _ConfigUpdater):
                attr(name, app)


def _force_notfound_prefix(app: Sphinx, _: Config) -> None:
    if "READTHEDOCS" not in os.environ:
        return

    if config_provided_by_user(app, "notfound_urls_prefix"):
        return

    components = urllib.parse.urlparse(os.environ["READTHEDOCS_CANONICAL_URL"])
    app.config.notfound_urls_prefix = components.path


def setup(app: Sphinx) -> dict[str, Any]:
    """Set up rocm_docs.core as a Sphinx extension."""
    required_extensions = [
        "myst_nb",
        "notfound.extension",
        "rocm_docs.projects",
        "sphinx_copybutton",
        "sphinx_design",
        "sphinx.ext.autodoc",
        "sphinx.ext.autosummary",
        "sphinx.ext.doctest",
        "sphinx.ext.duration",
    ]
    for ext in required_extensions:
        app.setup_extension(ext)

    app.add_config_value(
        "setting_all_article_info", default=False, rebuild="html", types=str
    )
    app.add_config_value(
        "all_article_info_os",
        default=[],
        rebuild="html",
        types=str,
    )
    app.add_config_value(
        "all_article_info_author", default="", rebuild="html", types=str
    )
    app.add_config_value(
        "all_article_info_date", default="", rebuild="html", types=str
    )
    app.add_config_value(
        "all_article_info_read_time", default="", rebuild="html", types=str
    )
    app.add_config_value(
        "article_pages", default=[], rebuild="html", types=list
    )

    # Run before notfound.extension sees the config (default priority(=500))
    app.connect("config-inited", _force_notfound_prefix, priority=400)
    app.connect("config-inited", _DefaultSettings.update_config)
    app.connect("build-finished", article_info.set_article_info, priority=1000)
    return {"parallel_read_safe": True, "parallel_write_safe": True}



"""Defines the ROCmDocs package.

Set up variables for documentation of ROCm projects
that are using Read the Docs.
"""

from __future__ import annotations

from typing import ClassVar, TypeAlias

import os

from rocm_docs.core import setup

MaybePath: TypeAlias = str | os.PathLike[str] | None


# Intentionally disabling the too-many-instance-attributes check in pylint
# as this class is intended to contain all necessary Sphinx config variables
# pylint: disable=too-many-instance-attributes
class ROCmDocs:
    """A class to contain all of the Sphinx variables."""

    SPHINX_VARS: ClassVar = [
        "extensions",
        "html_title",
        "html_theme",
        "html_theme_options",
        "doxygen_root",
        "doxygen_project",
        "doxyfile",
        "doxysphinx_enabled",
    ]

    def __init__(
        self,
        project_name: str,
        _: str | None = None,
        __: MaybePath = None,
    ) -> None:
        """Intialize ROCmDocs."""
        self._project_name: str = project_name
        self.extensions: list[str] = []
        self.html_title: str
        self.html_theme: str
        self.html_theme_options: dict[str, str | (bool | list[str])] = {}
        self.doxygen_root: MaybePath = None
        self.doxygen_project: dict[str, str | None | MaybePath] = {
            "name": None,
            "path": None,
        }
        self.doxyfile: MaybePath = None
        self.doxysphinx_enabled: bool = False

    @property
    def project(self) -> str:
        """Sphinx project variable."""
        return self._project_name

    def run_doxygen(
        self,
        doxygen_root: MaybePath = None,
        doxygen_path: MaybePath = None,
        doxygen_file: str | None = None,
    ) -> None:
        """Run doxygen as part of Sphinx by adding rocm_docs.doxygen."""
        if "rocm_docs.doxygen" not in self.extensions:
            self.extensions.append("rocm_docs.doxygen")

        self.doxygen_root = doxygen_root
        self.doxygen_project = {
            "name": self._project_name,
            "path": doxygen_path,
        }
        self.doxyfile = doxygen_file

    def enable_api_reference(self) -> None:
        """Enable embedding the doxygen generated api."""
        if "rocm_docs.doxygen" not in self.extensions:
            self.extensions.append("rocm_docs.doxygen")

        self.doxysphinx_enabled = True

    def setup(self) -> None:
        """Set up default RTD variables."""
        self.extensions.append("rocm_docs")
        full_project_name = self._project_name
        self.html_title = full_project_name
        self.html_theme = "rocm_docs_theme"


__all__ = ["ROCmDocs", "setup"]



"""Handle external projects.

Remote loading of intersphinx_mapping from file, templating projects in toc.yml).
"""

from __future__ import annotations

from typing import Any, TypeAlias, cast

import functools
import importlib.resources
import json
import os
import re
import sys
import urllib.parse
from dataclasses import dataclass
from pathlib import Path

import fastjsonschema  # type: ignore[import-untyped]
import github
import requests
import sphinx.util.logging
import yaml
from packaging.version import Version
from pydata_sphinx_theme.utils import (  # type: ignore[import-untyped]
    config_provided_by_user,
)
from sphinx.application import Sphinx
from sphinx.config import Config
from sphinx.errors import ExtensionError

from rocm_docs import formatting, theme, util

if sys.version_info < (3, 11):
    import importlib.abc as importlib_abc
else:
    import importlib.resources.abc as importlib_abc

Traversable = importlib_abc.Traversable

Inventory: TypeAlias = str | None | tuple[str | None, ...]
ProjectMapping: TypeAlias = tuple[str, Inventory]

DEFAULT_INTERSPHINX_REPOSITORY = "ROCm/rocm-docs-core"
DEFAULT_INTERSPHINX_BRANCH = "develop"
DOCS_VERSION_PATTERN = r"^docs-\d+\.\d+\.\d+$"

logger = sphinx.util.logging.getLogger(__name__)


class InvalidMappingFileError(RuntimeError):
    """Mapping file has invalid format, or failed to validate."""


ProjectItem: TypeAlias = str | list[str | None] | None
ProjectEntry: TypeAlias = str | dict[str, ProjectItem]


@dataclass
class _Project:
    target: str
    inventory: list[str | None]
    development_branch: str
    valid_versions: list[str | None]
    doxygen_html: str | None = None

    @staticmethod
    @functools.lru_cache
    def yaml_schema() -> dict[str, Any]:
        base = importlib.resources.files("rocm_docs") / "data"
        schema_file = base / "projects.schema.json"

        return cast(
            dict[str, Any], json.load(schema_file.open(encoding="utf-8"))
        )

    @classmethod
    def schema(cls) -> dict[str, Any]:
        return cast(dict[str, Any], cls.yaml_schema()["$defs"]["project"])

    @classmethod
    def default_value(cls, prop: str) -> str:
        return cast(str, cls.schema()["properties"][prop]["default"])

    @staticmethod
    def _get_doxygen_html(entry: ProjectEntry) -> str | None:
        assert isinstance(entry, dict)

        if "doxygen" not in entry:
            return None

        doxygen_entry = entry["doxygen"]
        assert isinstance(doxygen_entry, dict | str)

        if isinstance(doxygen_entry, dict):  # type:ignore
            doxygen_entry = doxygen_entry["html"]  # type:ignore

        # Parse as a URI, but only allow the path component
        urlparts = urllib.parse.urlsplit(doxygen_entry)
        for key, value in urlparts._asdict().items():
            if key == "path" or value == "":
                continue
            raise ExtensionError(
                f"URL not allowed for 'doxygen': {doxygen_entry}"
            )
        if urlparts.path.startswith("/"):
            raise ExtensionError(
                f"URL for 'doxygen' must be relative: {doxygen_entry}"
            )

        return urlparts.path

    @classmethod
    def from_yaml_entry(cls, entry: ProjectEntry) -> _Project:
        """Create from an entry that conforms to the project schema."""
        if isinstance(entry, str):
            return _Project(
                entry,
                [cls.default_value("inventory")],
                cls.default_value("development_branch"),
                [],  # No valid_versions for simple string entries
            )

        # It's okay to just index into optional fields, because jsonschema
        # fills in any missing fields with their default values
        inventory = entry["inventory"]
        assert inventory is None or isinstance(inventory, list | str)
        if not isinstance(inventory, list):
            inventory = [inventory]

        valid_versions = entry.get("valid_versions", [])
        assert valid_versions is None or isinstance(valid_versions, list | str)
        if not isinstance(valid_versions, list):
            valid_versions = (
                [valid_versions] if valid_versions is not None else []
            )

        return _Project(
            cast(str, entry["target"]),
            inventory,
            cast(str, entry["development_branch"]),
            valid_versions,
            cls._get_doxygen_html(entry),
        )

    @classmethod
    def get_static_version(
        cls,
        current_branch: str,
        current_project: _Project | None,
    ) -> str | None:
        """Returns a common static version if it exists.

        In some cases all remote projects will receive the same version,
        return that version if this is the case, returns None otherwise.
        """
        # Canonically available everywhere
        if current_branch in ["latest", "stable"]:
            return current_branch

        # Past release versions always with docs/
        regex = re.compile(DOCS_VERSION_PATTERN)
        if regex.match(current_branch):
            return current_branch

        # Anything besides the canonical development branch links to latest docs
        development_branch: str = cls.default_value("development_branch")
        if current_project is not None:
            development_branch = current_project.development_branch

        if current_branch != development_branch:
            return "latest"

        return None

    def _find_best_valid_version(self, target_version: str) -> str:
        """Find the best matching valid version for this project.

        If valid_versions is empty, return the target_version as-is.
        Otherwise, find the closest valid version.
        """
        if not self.valid_versions:
            return target_version

        # If target version is in valid versions, use it directly
        if target_version in self.valid_versions:
            return target_version

        # For specific docs- versions, find the closest valid version
        regex = re.compile(DOCS_VERSION_PATTERN)
        if regex.match(target_version):
            doc_versions = [
                version
                for version in self.valid_versions
                if version and regex.match(version)
            ]
            if doc_versions:
                try:

                    def calc_version(version_str: str) -> Version:
                        return Version(version_str.replace("docs-", ""))

                    target_version_value = calc_version(target_version)
                    valid_versions = [
                        (calc_version(v), v) for v in doc_versions
                    ]
                    valid_versions.sort(reverse=True)  # sort versions

                    # Find first version <= target
                    for version_value, version_string in valid_versions:
                        if version_value <= target_version_value:
                            return version_string

                    # Return last valid version, if target version is smaller
                    # than any valid versions.
                    return valid_versions[-1][1]
                except ValueError:
                    # If we can't parse the version, use the the target_version
                    # as-is.
                    pass

        # Fallback: return the target_version as-is.
        return target_version

    def evaluate(self, static_version: str | None) -> None:
        """Evaluate ${version} placeholders in the inventory and target values.

        Edge case:
        For docs/a.b.c versions/branches, ReadtheDocs replaces the / with -
        So handle GitHub version and RTD version differently
        """
        version = (
            static_version
            if static_version is not None
            else self.development_branch
        )

        # Apply valid_versions filtering
        version = self._find_best_valid_version(version)

        gh_version = version

        # edge case
        if version.startswith("docs-"):
            gh_version = version.replace("-", "/")

        if "${version}" in self.target:
            self.target = self.target.replace("${version}", version)
        elif "${gh_version}" in self.target:
            self.target = self.target.replace("${gh_version}", gh_version)

        for i, item in enumerate(self.inventory):
            if item is None:
                continue
            if "${version}" in item:
                self.inventory[i] = item.replace("${version}", version)
            elif "${gh_version}" in item:
                self.inventory[i] = item.replace("${gh_version}", gh_version)

    @property
    def mapping(self) -> ProjectMapping:
        """Target and inventory location in the format expected by sphinx."""
        return (self.target, tuple(self.inventory))


def _create_projects(project_yaml: str | Traversable) -> dict[str, _Project]:
    contents = yaml.safe_load(
        project_yaml
        if isinstance(project_yaml, str)
        else project_yaml.open(encoding="utf-8")
    )

    data: dict[str, int | dict[str, ProjectEntry]]
    try:
        data = fastjsonschema.validate(_Project.yaml_schema(), contents)
    except fastjsonschema.exceptions.JsonSchemaValueException as err:
        raise InvalidMappingFileError(
            f"Mapping file is invalid: {err.message}."
        ) from err

    assert isinstance(data["projects"], dict)
    return {
        project_id: _Project.from_yaml_entry(entry)
        for project_id, entry in data["projects"].items()
    }


def _get_current_project(
    projects: dict[str, _Project], current_id: str
) -> _Project | None:
    if current_id in projects:
        return projects[current_id]

    logger.warning(
        f"Current project '{current_id}' not found in projects.\n"
        "Did you forget to set 'external_projects_current_project' to "
        "the name of the current project?"
    )
    return None


def _create_mapping(
    projects: dict[str, _Project],
    current_project: _Project | None,
    current_branch: str,
) -> dict[str, ProjectMapping]:
    static_version = _Project.get_static_version(
        current_branch, current_project
    )
    for project in projects.values():
        project.evaluate(static_version)

    return {name: project.mapping for name, project in projects.items()}


class MappingFileFetchError(RuntimeError):
    """Fetching the yaml file from the remote failed."""


def _fetch_projects(
    remote_repository: str,
    remote_branch: str,
    remote_filepath: str,
) -> str:
    try:
        gh_api = github.Github(os.environ.get("TOKEN"))
        repo = gh_api.get_repo(remote_repository)
        contents = repo.get_contents(remote_filepath, remote_branch)
        if isinstance(contents, list):
            raise MappingFileFetchError("Expected a file not a directory!")

        return contents.decoded_content.decode("utf-8")
    except github.GithubException as err:
        assert isinstance(err.data["message"], str)
        message: str = err.data["message"]
        raise MappingFileFetchError(
            "failed to read remote mappings from "
            f"{remote_repository}:{remote_filepath} "
            f"on branch={remote_branch}, API returned {err.status}: {message}.",
        ) from err


def _load_projects(
    remote_repository: str, remote_branch: str
) -> dict[str, _Project]:
    projects_file_loc = "data/projects.yaml"

    def should_fetch_mappings(
        remote_repository: str | None, remote_branch: str | None
    ) -> bool:
        if not remote_repository:
            logger.info(
                "Skipping the fetch for remote mappings, remote_repository "
                "is unset."
            )
            return False

        if not remote_branch:
            logger.error(
                "Remote branch is unset, cannot fetch remote mappings."
            )
            return False

        logger.info(
            "Remote mappings will be fetched from "
            f"{remote_repository} branch={remote_branch}"
        )
        return True

    projects: dict[str, _Project] | None = None
    if should_fetch_mappings(remote_repository, remote_branch):
        try:
            remote_filepath = "src/rocm_docs/" + projects_file_loc
            projects = _create_projects(
                _fetch_projects(
                    remote_repository,
                    remote_branch,
                    remote_filepath,
                )
            )
        except (MappingFileFetchError, InvalidMappingFileError) as err:
            logger.warning(
                f"Failed to use remote mapping: {err} "
                "Falling back to bundled mapping."
            )

    if projects is None:
        projects = _create_projects(
            importlib.resources.files("rocm_docs") / projects_file_loc
        )

    return projects


def _get_context(
    repo_path: Path, mapping: dict[str, ProjectMapping]
) -> dict[str, Any]:
    url, branch = util.get_branch(repo_path)
    return {
        "url": url,
        "branch": branch,
        "projects": {k: v[0].rstrip("/") for k, v in mapping.items()},
    }


def _update_theme_configs(
    app: Sphinx,
    current_project: _Project | None,
    current_branch: str,
    flavor: str,
) -> None:
    """Update configurations for use in theme.py"""
    latest_version_list = requests.get(
        "https://raw.githubusercontent.com/ROCm/rocm-docs-core/new_data/latest_version.txt"
    ).text.strip()
    latest_version_dict = theme._parse_version(latest_version_list)
    latest_version = latest_version_dict.get(flavor, "latest")
    latest_version_string_list = ["latest"]
    if latest_version != "latest":
        # Some component's docs branch has "docs-" prefix, others do not
        latest_version_string_list += [f"docs-{latest_version}", latest_version]

    release_candidate = requests.get(
        "https://raw.githubusercontent.com/ROCm/rocm-docs-core/new_data/release_candidate.txt"
    ).text.strip("\r\n")
    release_candidate_string = f"docs-{release_candidate}"

    development_branch = _Project.default_value("development_branch")
    if current_project is not None:
        development_branch = current_project.development_branch

    doc_branch_pattern = r"^docs-\d+\.\d+\.\d+$"

    if flavor == "rocm" and current_branch in latest_version_string_list:
        app.config.projects_version_type = util.VersionType.ROCM_LATEST_RELEASE
    elif flavor != "rocm" and current_branch in latest_version_string_list:
        app.config.projects_version_type = util.VersionType.OTHER_LATEST_RELEASE
    elif current_branch.endswith("preview"):
        app.config.projects_version_type = util.VersionType.PREVIEW
    elif current_branch.startswith(release_candidate_string):
        app.config.projects_version_type = util.VersionType.RELEASE_CANDIDATE
    elif re.match(doc_branch_pattern, current_branch):
        app.config.projects_version_type = util.VersionType.OLD_RELEASE
    elif current_branch == development_branch:
        app.config.projects_version_type = util.VersionType.DEVELOPMENT


def _get_external_projects(
    app: Sphinx, default: dict[str, ProjectMapping]
) -> list[str]:
    external_projects: list[str] | str = app.config.external_projects
    if external_projects == "all":
        return list(default.keys())
    if isinstance(external_projects, str):
        logger.error(
            f'Unexpected value "{external_projects}" in external_projects.\n'
            'Must be set to a list of project names or "all" to '
            "enable all projects defined in projects.yaml"
        )
        return []

    unknown_projects = [p for p in external_projects if p not in default]
    if len(unknown_projects) > 0:
        known_projects = [f'"{p}"' for p in default]
        unknown_projects = [f'"{p}"' for p in unknown_projects]
        logger.error(
            "Unknown projects: [{}] in external_projects.\n".format(
                ", ".join(unknown_projects)
            )
            + "Valid projects are: [{}]".format(", ".join(known_projects))
        )
    return external_projects


def _set_doxygen_html(app: Sphinx, current_project: _Project | None) -> None:
    if current_project is None or current_project.doxygen_html is None:
        return

    if not hasattr(app.config, "doxygen_html"):
        return

    doxygen_html = current_project.doxygen_html
    if config_provided_by_user(app, "doxygen_html"):
        if doxygen_html != app.config.doxygen_html:
            logger.warning(
                f'The setting doxygen_html="{app.config.doxygen_html}"'
                f' differs from projects.yaml value: "{doxygen_html}"'
            )
        return

    app.config.doxygen_html = doxygen_html


def _update_config(app: Sphinx, _: Config) -> None:
    if not config_provided_by_user(app, "intersphinx_disabled_domains"):
        app.config.intersphinx_disabled_domains = ["std"]

    remote_repository = app.config.external_projects_remote_repository
    remote_branch = app.config.external_projects_remote_branch
    projects = _load_projects(remote_repository, remote_branch)

    repo_path = Path(app.srcdir)
    __, branch = util.get_branch(repo_path)
    current_project = _get_current_project(
        projects, app.config.external_projects_current_project
    )
    remote_mapping = _create_mapping(projects, current_project, branch)
    external_projects = _get_external_projects(app, remote_mapping)

    mapping: dict[str, ProjectMapping] = app.config.intersphinx_mapping
    for key, value in remote_mapping.items():
        if key in external_projects:
            mapping.setdefault(key, value)

    if not config_provided_by_user(app, "external_toc_path"):
        app.config.external_toc_path = "./.sphinx/_toc.yml"

    context = _get_context(Path(app.srcdir), remote_mapping)
    formatting.format_toc(
        Path(app.srcdir, app.config.external_toc_template_path),
        Path(app.srcdir, app.config.external_toc_path),
        context,
    )

    if not config_provided_by_user(app, "html_theme_options"):
        app.config.html_theme_options = {"flavor": "rocm"}

    # Store the context to be referenced later
    app.config.projects_context = context

    _set_doxygen_html(app, current_project)
    _update_theme_configs(
        app,
        current_project,
        branch,
        app.config.html_theme_options.get("flavor", "rocm"),
    )


def _setup_projects_context(
    app: Sphinx, _: str, __: str, context: dict[str, Any], ___: Any
) -> None:
    context["projects"] = app.config.projects_context["projects"]


def setup(app: Sphinx) -> dict[str, Any]:
    """Setup rocm_docs.projects as a sphinx extension."""
    app.setup_extension("sphinx.ext.intersphinx")
    app.setup_extension("sphinx_external_toc")

    app.add_config_value(
        "external_projects_remote_repository",
        DEFAULT_INTERSPHINX_REPOSITORY,
        rebuild="env",
        types=str,
    )
    app.add_config_value(
        "external_projects_remote_branch",
        DEFAULT_INTERSPHINX_BRANCH,
        rebuild="env",
        types=str,
    )
    app.add_config_value(
        "external_projects_current_project",
        lambda config: config.project,
        rebuild="env",
        types=str,
    )
    app.add_config_value(
        "external_projects", "all", rebuild="env", types=[list, str]
    )

    def external_toc_template_default(config: Config) -> str:
        toc_path = Path(config.external_toc_path)
        return str(toc_path.with_suffix(toc_path.suffix + ".in"))

    app.add_config_value(
        "external_toc_template_path",
        external_toc_template_default,
        rebuild="env",
        types=[str, Path],
    )

    # This needs to happen before external-tocs's config-inited (priority=900)
    app.connect("config-inited", _update_config)
    app.connect("html-page-context", _setup_projects_context)
    return {"parallel_read_safe": True, "parallel_write_safe": True}


def debug_projects() -> None:
    """Get remote mappings display them and format the toc.

    Provided as a debugging tool for the functionality of this module.
    """
    projects = _load_projects(
        DEFAULT_INTERSPHINX_REPOSITORY, DEFAULT_INTERSPHINX_BRANCH
    )
    print(projects)

    current_project = _get_current_project(projects, "rocm-docs-core")
    print(current_project)

    repo_path = Path()
    _, branch = util.get_branch(repo_path)
    mapping = _create_mapping(projects, current_project, branch)
    print(mapping)
    context = _get_context(Path(), mapping)
    print(context)
    toc_in = Path("./.sphinx/_toc.yml.in")
    if len(sys.argv) > 1:
        toc_in = Path(sys.argv[1])
    toc_out = toc_in.with_suffix("")
    if len(sys.argv) > 2:
        toc_out = Path(sys.argv[2])

    formatting.format_toc(
        toc_in,
        toc_out,
        context,
    )


if __name__ == "__main__":
    debug_projects()



from rocm_docs import ROCmDocs

# base setup for using rocm-docs-core package
docs_core = ROCmDocs("ROCm Docs Core Doxygen Test Project - Legacy")
docs_core.setup()

# doxygen integration
docs_core.run_doxygen(doxygen_root="doxygen", doxygen_path="doxygen/xml")

# doxysphinx integration
docs_core.enable_api_reference()

# used in intersphinx linking
external_projects_current_project = "a"

# custom path to table of contents
external_toc_path = "./sphinx/_toc.yml"

# set variables from rocm-docs-core for sphinx
for sphinx_var in ROCmDocs.SPHINX_VARS:
    globals()[sphinx_var] = getattr(docs_core, sphinx_var)

# optional extension for additional linking to doxygen
if not "extensions" in globals():
    extensions = []
extensions += ["sphinxcontrib.doxylink"]



extensions = ["rocm_docs"]
html_theme = "rocm_docs_theme"
html_theme_options = {"flavor": "rocm-blogs"}

external_toc_path = "./sphinx/_toc.yml"

external_projects_current_project = "a"

version = "1.0.0"
release = "1.0.0"
html_title = f"ROCm Docs Core Theme Test Project - ROCm Blogs Flavor"
project = "ROCm Docs Core"
author = "Advanced Micro Devices, Inc."
copyright = (
    "Copyright (c) 2026 Advanced Micro Devices, Inc. All rights reserved."
)



"""Defines helpers for testing the rocm_docs sphinx extension"""

from __future__ import annotations

import functools
import shutil
from collections.abc import Callable, Iterator
from pathlib import Path

import pytest
from sphinx.application import Sphinx

from .log_fixtures import ExpectLogFixture


@pytest.fixture
def with_no_git_repo(
    monkeypatch: pytest.MonkeyPatch,
    expect_log: ExpectLogFixture,
) -> Iterator[ExpectLogFixture.Validator]:
    """Setup environment to allow testing outside a git repository"""
    monkeypatch.setenv("ROCM_DOCS_REMOTE_DETAILS", ",")

    with expect_log(
        "git.exc.InvalidGitRepositoryError",
        "ERROR",
        "test_external_projects",
    ) as validator:
        yield validator


SITES_BASEFOLDER = Path(__file__).parent / "sites"


def build_sphinx(
    srcdir: Path, outdir: Path, confdir: Path | None = None
) -> None:
    confdir = confdir or srcdir
    doctreedir = outdir / ".doctrees"
    buildername = "html"
    app = Sphinx(srcdir, confdir, outdir, doctreedir, buildername)
    app.build()


@pytest.fixture
def build_factory(
    request: pytest.FixtureRequest,
    tmp_path: Path,
) -> Callable[..., tuple[Path, Path]]:
    """A factory to prepare Sphinx source and output directories"""

    def make(src_folder: Path, /) -> tuple[Path, Path]:
        srcdir = tmp_path.joinpath(src_folder)
        outdir = tmp_path.joinpath(f"{src_folder}_build")
        srcdir.parent.mkdir(parents=True, exist_ok=True)

        mark = request.node.get_closest_marker("template_folder")
        if mark is not None:
            for i, dir in enumerate(mark.args):
                shutil.copytree(dir, srcdir, dirs_exist_ok=(i > 0))

        shutil.copytree(
            SITES_BASEFOLDER / src_folder, srcdir, dirs_exist_ok=True
        )
        return srcdir, outdir

    if hasattr(request, "param"):
        return functools.partial(make, request.param)
    return make


__all__ = ["build_factory", "with_no_git_repo"]



"""Logic to add article info to a page.

For all options see the user guide:
https://rocm.docs.amd.com/projects/rocm-docs-core/en/latest/user_guide/article_info.html
"""

from typing import Any, cast

import importlib.resources
import os
import re
from pathlib import Path

import bs4
import git.repo
from sphinx.application import Sphinx
from sphinx.config import Config


def set_article_info(app: Sphinx, _: Config) -> None:
    """Add article info headers to HTML pages."""
    if (
        app.config.setting_all_article_info is False
        and len(app.config.article_pages) == 0
    ):
        return

    article_info = (
        importlib.resources.files("rocm_docs")
        .joinpath("rocm_docs_theme/components/article-info.html")
        .read_text(encoding="utf-8")
    )

    specific_pages: list[str] = []

    _set_page_article_info(app, article_info, specific_pages)

    if app.config.setting_all_article_info is True:
        _set_all_article_info(app, article_info, specific_pages)


def _set_page_article_info(
    app: Sphinx, article_info: str, specific_pages: list[str]
) -> None:
    """Add article info headers to the configured HTML pages.

    The pages can be set in "article_pages" of the Sphinx configuration.
    """
    repo = git.repo.Repo(app.srcdir, search_parent_directories=True)
    for page in app.config.article_pages:
        path_rel = app.project.doc2path(page["file"], False)
        path_html = Path(app.outdir, path_rel).with_suffix(".html")
        path_source = Path(app.srcdir, path_rel)

        # FIXME: This will silently skip all files when not building the default
        # `html` format (e.g `htmlzip`, `epub` or `pdf`)
        if not path_html.is_file():
            continue

        os_list = []
        page.setdefault("os", app.config.all_article_info_os)
        if "linux" in page["os"]:
            os_list.append("Linux")
        if "windows" in page["os"]:
            os_list.append("Windows")
        article_os_info = " and ".join(os_list)
        if os_list:
            article_os_info = f"Applies to {article_os_info}"
        modified_info = article_info.replace("<!--os-info-->", article_os_info)

        author = app.config.all_article_info_author
        if "author" in page:
            author = page["author"]
        modified_info = modified_info.replace("<!--author-info-->", author)

        date_info: str | None = None
        if "date" in page:
            date_info = page["date"]
        else:
            date_info = _get_time_last_modified(repo, path_source)

        if date_info == "":
            soup = bs4.BeautifulSoup(modified_info, "html.parser")
            svg_to_remove = soup.find("span", class_="article-info-date-svg")
            if svg_to_remove and isinstance(svg_to_remove, bs4.Tag):
                svg_to_remove.decompose()
            modified_info = str(soup)

        if date_info is not None:
            modified_info = modified_info.replace("<!--date-info-->", date_info)

        if "read-time" in page:
            read_time = page["read-time"]
        else:
            read_time = _estimate_read_time(path_html)

        if read_time == "":
            soup = bs4.BeautifulSoup(modified_info, "html.parser")
            svg_to_remove = soup.find(
                "span", class_="article-info-read-time-svg"
            )
            if svg_to_remove and isinstance(svg_to_remove, bs4.Tag):
                svg_to_remove.decompose()
            modified_info = str(soup)

        if read_time is not None:
            modified_info = modified_info.replace("<!--read-info-->", read_time)

        specific_pages.append(page["file"])
        _write_article_info(path_html, modified_info)


def _set_all_article_info(
    app: Sphinx, article_info: str, specific_pages: list[str]
) -> None:
    """Add article info headers with general settings to all HTML pages.

    Pages that have specific settings (configured by "article_pages") are
    skipped.
    """
    repo = git.repo.Repo(app.srcdir, search_parent_directories=True)
    for docname in app.project.docnames:
        # skip pages with specific settings
        if docname in specific_pages:
            continue

        page_rel = app.project.doc2path(docname, False)
        page = Path(app.outdir, page_rel).with_suffix(".html")

        # FIXME: This will silently skip all files when not building the default
        # `html` format (e.g `htmlzip`, `epub` or `pdf`)
        if not page.is_file():
            continue

        os_list = []
        if "linux" in app.config.all_article_info_os:
            os_list.append("Linux")
        if "windows" in app.config.all_article_info_os:
            os_list.append("Windows")
        article_os_info = " and ".join(os_list)
        if os_list:
            article_os_info = f"Applies to {article_os_info}"

        date_info = _get_time_last_modified(repo, Path(app.srcdir, page_rel))
        if not date_info:
            date_info = cast(str, app.config.all_article_info_date)

        modified_info = article_info.replace("<!--os-info-->", article_os_info)
        modified_info = modified_info.replace(
            "<!--author-info-->", app.config.all_article_info_author
        )
        modified_info = modified_info.replace("<!--date-info-->", date_info)
        modified_info = modified_info.replace(
            "<!--read-info-->", _estimate_read_time(page)
        )

        _write_article_info(page, modified_info)


def _get_time_last_modified(repo: git.repo.Repo, path: Path) -> str | None:
    try:
        time = next(
            repo.iter_commits(paths=path, max_count=1)
        ).committed_datetime
        return time.strftime("%Y-%m-%d")
    except StopIteration:
        return None


def _estimate_read_time(file_name: Path) -> str:
    def is_visible(element):
        if element.parent.name in [
            "style",
            "script",
            "[document]",
            "head",
            "title",
        ]:
            return False
        if isinstance(element, bs4.element.Comment):
            return False
        return element.string != "\n"

    def count_words(text):
        words = re.findall(r"\w+", text)
        return len(words)

    words_per_minute = 200

    with open(file_name, encoding="utf-8") as file:
        html = file.read()
    soup = bs4.BeautifulSoup(html, "html.parser")
    page_text = soup.find_all(text=True)
    visible_page_text = filter(is_visible, page_text)
    average_word_count = sum(count_words(line) for line in visible_page_text)
    time_minutes = int(max(1, round(average_word_count / words_per_minute)))
    return f"{time_minutes} min read time"


def _write_article_info(path: os.PathLike[Any], article_info: str) -> None:
    with open(path, "r+", encoding="utf8") as file:
        page_html = file.read()
        soup = bs4.BeautifulSoup(page_html, "html.parser")

        has_article_info = soup.find("div", id="rocm-docs-core-article-info")
        if (
            has_article_info is not None
            or soup.article is None
            or soup.article.h1 is None
        ):
            return

        soup.article.h1.insert_after(
            bs4.BeautifulSoup(article_info, "html.parser")
        )
        file.seek(0)
        file.truncate(0)
        file.write(str(soup))



from __future__ import annotations

from types import SimpleNamespace

import copy
import unittest.mock

import pytest

import rocm_docs.doxygen


def test_copy_tagfile(tmp_path_factory: pytest.TempPathFactory) -> None:
    app = unittest.mock.NonCallableMock()

    srcdir = tmp_path_factory.mktemp("srcdir")
    srcfile = srcdir.joinpath("tagfile.xml")
    expected_content = "Dummy doxygen tagfile"
    srcfile.write_text(expected_content)

    outdir = tmp_path_factory.mktemp("outdir").absolute()
    outfile = outdir.joinpath("tagfile.xml")

    app.config.doxygen_html = ""
    app.builder = SimpleNamespace(format="html")
    app.outdir = str(outdir)
    app.srcdir = str(srcdir)

    rocm_docs.doxygen._copy_tagfile(app)

    assert outfile.read_text() == expected_content


@pytest.mark.parametrize(
    ("output_format", "doxygen_html"),
    [("html", None), ("epub", "")],
    ids=["not_set", "skipped_format"],
)
def test_copy_tagfile_skipped(
    tmp_path_factory: pytest.TempPathFactory,
    output_format: str,
    doxygen_html: str | None,
) -> None:
    app = unittest.mock.NonCallableMock()

    srcdir = tmp_path_factory.mktemp("srcdir")
    srcfile = srcdir.joinpath("tagfile.xml")
    srcfile.touch(exist_ok=False)

    outdir = tmp_path_factory.mktemp("outdir").absolute()
    outfile = outdir.joinpath("tagfile.xml")

    app.config.doxygen_html = doxygen_html
    app.builder = SimpleNamespace(format=output_format)
    app.outdir = str(outdir)
    app.srcdir = str(srcdir)

    rocm_docs.doxygen._copy_tagfile(app)

    assert not outfile.exists()


@pytest.mark.parametrize(
    ("has_doxylink", "has_doxygen_html"),
    [(False, True), (True, False), (False, False)],
    ids=["no_doxylink", "no_html", "neither"],
)
def test_update_doxylink_settings_skipped(
    has_doxylink: bool, has_doxygen_html: bool
) -> None:
    app = unittest.mock.NonCallableMock(spec=["config", "srcdir"])
    app.srcdir = ""
    config_spec = ["doxygen_html", "overrides", "_raw_config"]

    if has_doxylink:
        config_spec.append("doxylink")

    config = unittest.mock.NonCallableMock(config_spec)
    config.doxygen_html = "doxygen/html" if has_doxygen_html else None
    config.overrides = []
    config._raw_config = []
    expected: dict[str, tuple[str, str]]
    if has_doxylink:
        expected = {"must-not-be-changed": ("tagfile.xml", "test-string")}
        config.doxylink = copy.copy(expected)
    app.config = config

    rocm_docs.doxygen._update_doxylink_settings(app, config)

    if has_doxylink:
        assert app.config.doxylink == expected
    else:
        assert not hasattr(app.config, "doxylink")



import yaml
from typing import List, Dict, Any
import sys

class ListFlowStyleRepresenter:
    """
    Custom representer to force specific flow style for nested lists
    """
    def __init__(self):
        self.add_representer()

    def represent_list(self, dumper, data):
        # Check if this is a nested list containing strings/patterns
        if data and isinstance(data[0], str):
            return dumper.represent_sequence('tag:yaml.org,2002:seq', data, flow_style=True)
        return dumper.represent_sequence('tag:yaml.org,2002:seq', data, flow_style=False)

    def add_representer(self):
        yaml.add_representer(list, self.represent_list)

def load_yaml(file_path: str) -> Any:
    """
    Load YAML file with error handling.
    """
    try:
        with open(file_path, 'r', encoding='utf-8') as f:
            return yaml.safe_load(f)
    except yaml.scanner.ScannerError as e:
        print(f"\nYAML Syntax Error in {file_path}:")
        print(f"Line {e.problem_mark.line + 1}, Column {e.problem_mark.column + 1}")
        print(f"Error: {e.problem}")
        return None
    except Exception as e:
        print(f"\nError loading {file_path}:")
        print(str(e))
        return None

def merge_matrix_entries(default_entries: List[Dict], user_entries: List[Dict]) -> List[Dict]:
    """
    Merge matrix entries, combining sources for entries with matching names.

    Supports merge modes via '_merge_mode' key in user entries:
    - 'overwrite': Replace default sources entirely
    - 'append': Add to the end of default sources (default behavior)

    Skip merging if the 'sources' list in the user entries is empty.
    """
    result = default_entries.copy()
    default_map = {entry.get('name'): entry for entry in result}

    for user_entry in user_entries:
        user_name = user_entry.get('name')
        if user_name and user_name in default_map:
            if 'sources' in user_entry:
                user_sources = user_entry['sources']

                # Get merge mode (default to 'append')
                merge_mode = user_entry.get('_merge_mode', 'append')

                # Skip merging if user sources are empty
                if not user_sources or all(not source for source in user_sources):
                    continue

                # Check the main configuration's sources
                default_sources = default_map[user_name].get('sources', [])

                # If the main config's sources are empty, replace them
                if not default_sources or all(not source for source in default_sources):
                    default_map[user_name]['sources'] = user_sources
                elif merge_mode == 'overwrite':
                    # Overwrite mode: replace entirely
                    default_map[user_name]['sources'] = user_sources
                    print(f"  → Overwriting sources for '{user_name}'")
                else:
                    # Default append mode
                    default_map[user_name]['sources'].extend(user_sources)
                    print(f"  → Appending sources for '{user_name}'")

                # Remove the _merge_mode key from the final output if present
                if '_merge_mode' in default_map[user_name]:
                    del default_map[user_name]['_merge_mode']

    return result

def merge_configs(default: Any, user: Any) -> Any:
    """
    Recursively merge two configurations.
    """
    if user is None:
        return default

    if isinstance(default, dict) and isinstance(user, dict):
        result = default.copy()
        for key, value in user.items():
            if key in result:
                result[key] = merge_configs(result[key], value)
            else:
                result[key] = value
        return result

    if isinstance(default, list) and isinstance(user, list):
        if (len(default) > 0 and isinstance(default[0], dict) and
            'name' in default[0] and 'sources' in default[0]):
            return merge_matrix_entries(default, user)
        return default + user

    return user

def save_yaml(data: Dict, file_path: str) -> bool:
    """
    Save YAML file with custom formatting.
    """
    try:
        # Initialize custom representer
        ListFlowStyleRepresenter()

        with open(file_path, 'w', encoding='utf-8') as f:
            yaml.dump(data, f, default_flow_style=False, sort_keys=False, allow_unicode=True)
        return True
    except Exception as e:
        print(f"\nError saving {file_path}:")
        print(str(e))
        return False

if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description='Merge spellcheck YAML configurations')
    parser.add_argument('template', help='Path to template YAML configuration')
    parser.add_argument('local', help='Path to local YAML configuration')
    parser.add_argument('output', help='Path to save merged configuration')

    args = parser.parse_args()

    print(f"\nMerging YAML configurations...")
    print(f"Template: {args.template}")
    print(f"Local:    {args.local}")
    print(f"Output:   {args.output}")
    print()

    # Load configurations
    template_config = load_yaml(args.template)
    if template_config is None:
        sys.exit(1)

    local_config = load_yaml(args.local)
    if local_config is None:
        sys.exit(1)

    # Merge configurations
    merged_config = merge_configs(template_config, local_config)

    # Save merged configuration
    if not save_yaml(merged_config, args.output):
        sys.exit(1)

    print(f"\nSuccessfully merged configurations into {args.output}")



"""Doxygen sub-extension of rocm-docs-core."""

from __future__ import annotations

from typing import Any, Union, cast

import importlib.metadata
import importlib.resources
import importlib.util
import os
import shutil
import subprocess
import sys
from pathlib import Path

from pydata_sphinx_theme.utils import (  # type: ignore[import-untyped]
    config_provided_by_user,
)
from sphinx.application import Sphinx
from sphinx.config import Config
from sphinx.errors import ConfigError, ExtensionError
from sphinx.util import logging
from sphinx.util.display import progress_message
from sphinx.util.osutil import copyfile

logger = logging.getLogger(__name__)


def _copy_files(app: Sphinx) -> None:
    """Insert additional files into workspace."""
    with importlib.resources.path("rocm_docs", "") as pkg_path:
        shutil.copytree(
            pkg_path / "data/_doxygen",
            Path(app.srcdir, "_doxygen"),
            dirs_exist_ok=True,
        )


def _get_config_default(config: Config, key: str) -> Any:
    """Get the default value for a sphinx configuration key."""
    default_or_callable = config.values[key].default
    if callable(default_or_callable):
        return default_or_callable(config)
    return default_or_callable


def _run_doxygen(app: Sphinx, config: Config) -> None:
    """Run doxygen, validating all paths."""
    # To support the (legacy) ROCmDocs interface 'None' is a synonym for the
    # default value
    for i in ["doxygen_root", "doxyfile"]:
        if getattr(config, i) is None:
            setattr(config, i, _get_config_default(config, i))

    doxygen_root = Path(app.confdir, config.doxygen_root)
    try:
        doxygen_root = doxygen_root.resolve(strict=True)
    except FileNotFoundError as err:
        raise ConfigError(
            f"Expected doxygen root folder {doxygen_root} to exist"
            " and be readable."
        ) from err

    if config["doxygen_executable"] is None:
        cmd_path = shutil.which("doxygen")
        if cmd_path is None:
            raise RuntimeError(
                "'doxygen' command not found! Make sure that "
                "doxygen is installed and in the PATH or specify "
                "via doxygen_executable configuration variable."
            )
        doxygen_exe = Path(cmd_path)
    else:
        doxygen_exe = Path(app.confdir, config.doxygen_executable)

    doxyfile = Path(app.confdir, config["doxyfile"]).absolute()
    if not doxyfile.is_file():
        raise ConfigError(
            f"Expected doxyfile {doxyfile} to exist and be readable."
        )

    # Running doxygen requires that the files are already copied because
    # the Doxyfile references files distributed with rocm-docs-core
    # (e.g. stylesheets)
    if app.config.doxysphinx_enabled:
        _copy_files(app)
    try:
        subprocess.check_call([doxygen_exe, "--version"], cwd=doxygen_root)
        subprocess.check_call([doxygen_exe, doxyfile], cwd=doxygen_root)
    except subprocess.CalledProcessError as err:
        raise RuntimeError("Failed when running doxygen") from err

    _update_breathe_settings(app, doxygen_root)
    _run_doxysphinx(app, doxygen_root, doxyfile, doxygen_exe)


def _update_breathe_settings(app: Sphinx, doxygen_root: Path) -> None:
    if config_provided_by_user(
        app, "breathe_projects"
    ) or config_provided_by_user(app, "breathe_default_project"):
        return

    doxygen_project: dict[str, None | str | os.PathLike[Any]] = (
        app.config.doxygen_project
    )

    # To support the (legacy) ROCmDocs interface 'None' is a synonym for the
    # default value for each element of the Tuple
    default: dict[str, None | str | os.PathLike[Any]] = _get_config_default(
        app.config, "doxygen_project"
    )
    for key, value in doxygen_project.items():
        if value is None:
            doxygen_project[key] = default[key]

    # FIXME: Log an error and reset to default instead of asserting
    assert isinstance(doxygen_project["name"], str)
    project_name: str = doxygen_project["name"]

    # First try relative to the
    assert doxygen_project["path"] is not None
    xml_path = Path(app.confdir, doxygen_project["path"])
    try:
        xml_path = xml_path.resolve(strict=True)
    except FileNotFoundError as err:
        raise NotADirectoryError(
            "Expected doxygen to generate the folder"
            f" {doxygen_root!s} but it could not be found."
        ) from err

    setattr(app.config, "breathe_projects", {project_name: str(xml_path)})
    setattr(app.config, "breathe_default_project", project_name)


def _update_doxylink_settings(app: Sphinx, _: Config) -> None:
    if not hasattr(app.config, "doxylink"):
        logger.info(
            "doxylink not enabled, skipping setting up the current" " project"
        )
        return

    if app.config.doxygen_html is None:
        return

    # Materialize the default value, since we're about to mutate it
    # Otherwise config.doxylink would return a temporary and any modification to
    # it would be lost.
    if not config_provided_by_user(app, "doxylink"):
        app.config.doxylink = app.config.doxylink

    doxylink = cast(
        dict[str, tuple[str, str] | tuple[str, str, str]],
        app.config.doxylink,
    )
    tagfile = Path(app.srcdir, app.config.doxygen_html, "tagfile.xml")

    doxylink.setdefault("doxygen", (str(tagfile), str(app.config.doxygen_html)))


def _run_doxysphinx(
    app: Sphinx, doxygen_root: Path, doxyfile: Path, doxygen_exe: Path
) -> None:
    if not app.config.doxysphinx_enabled:
        return

    if (
        "doxysphinx.cli" not in sys.modules
        and importlib.util.find_spec("doxysphinx.cli") is None
    ):
        raise RuntimeError(
            "Missing optional dependencies: make sure the "
            "[api_reference] feature is enabled. (e.g. "
            '"pip install rocm-docs-core[api_reference]")'
        )

    doxyphinx_version = importlib.metadata.version("doxysphinx")
    args = [
        sys.executable,
        "-m",
        "doxysphinx",
        "build",
        "--doxygen_exe=" + str(doxygen_exe.absolute()),
    ]
    if doxyphinx_version.endswith("+tagfile.toc"):
        args.append("--tagfile_toc")
    args += [str(app.srcdir), str(app.outdir), str(doxyfile)]

    try:
        subprocess.check_call(args, cwd=doxygen_root)
    except subprocess.CalledProcessError as err:
        raise RuntimeError(
            f"doxysphinx failed (exit code: {err.returncode})"
        ) from err


def _copy_tagfile(app: Sphinx) -> None:
    if app.config.doxygen_html is None:
        return

    if app.builder.format != "html":
        return

    src = Path(app.srcdir, app.config.doxygen_html, "tagfile.xml")
    dst = Path(app.outdir, src.name)
    with progress_message("copying doxygen tagfile"):
        try:
            copyfile(str(src), str(dst))
        except OSError as err:
            raise ExtensionError(f"Failed to copy tag file: {err}") from err


def setup(app: Sphinx) -> dict[str, Any]:
    """Set up rocm_docs.doxygen as a Sphinx extension."""
    app.setup_extension("sphinx.ext.mathjax")
    app.setup_extension("breathe")

    app.add_config_value(
        "doxygen_root", ".doxygen", rebuild="", types=[str, os.PathLike]
    )
    app.add_config_value(
        "doxygen_executable",
        None,
        rebuild="",
        types=[str, os.PathLike[Any]],
    )
    app.add_config_value(
        "doxyfile",
        lambda config: Path(config.doxygen_root, "Doxyfile"),
        rebuild="",
        types=[str, os.PathLike[Any]],
    )
    app.add_config_value(
        "doxygen_project",
        lambda config: {
            "name": "doxygen",
            "path": Path(config.doxygen_root, "docBin", "xml"),
        },
        rebuild="",
        types=dict[str, Union[str, "os.PathLike[Any]"]],
    )
    app.add_config_value("doxysphinx_enabled", False, rebuild="", types=bool)
    app.add_config_value("doxygen_html", None, rebuild="")

    # Should run before breathe sees their parameters, as we provide defaults.
    app.connect("config-inited", _run_doxygen, priority=400)
    # Has to run after projects.py config-inited, as it might set
    # doxygen_html
    app.connect("config-inited", _update_doxylink_settings, priority=500)
    # Should run after projects.py's config (if enabled) as it provides values
    # based on the contents projects.yaml, needs access to the builder
    app.connect("builder-inited", _copy_tagfile)

    return {"parallel_read_safe": True, "parallel_write_safe": True}



"""Module to use rocm-docs-core as a theme."""

from typing import Any

import time
from pathlib import Path

import requests
import sphinx.util.logging
from pydata_sphinx_theme.utils import (  # type: ignore[import-untyped]
    config_provided_by_user,
    get_theme_options_dict,
)
from sphinx.application import Sphinx

from rocm_docs import util

logger = sphinx.util.logging.getLogger(__name__)

MAX_RETRY = 100

ROCM_TOOLKITS_URL = "https://raw.githubusercontent.com/ROCm/rocm-docs-core/new_data/rocm_toolkits.txt"

ROCM_TOOLKITS_FALLBACK: dict[str, str] = {
    "ROCm Data Science": "https://rocm.docs.amd.com/projects/rocm-ds/en/latest/index.html",
    "ROCm Finance": "https://rocm.docs.amd.com/projects/rocm-finance/en/latest/index.html",
    "ROCm Life Science": "https://rocm.docs.amd.com/projects/rocm-ls/en/latest/index.html",
    "ROCm Simulation": "https://rocm.docs.amd.com/projects/rocm-simulation/en/latest/index.html",
}


def _get_rocm_toolkits() -> dict[str, str]:
    """Fetch rocm_toolkits.txt from new_data branch, retrying for up to 10 minutes.

    Falls back to ROCM_TOOLKITS_FALLBACK if the fetch ultimately fails.
    """
    deadline = time.monotonic() + 600  # 10 minutes
    headers = {"User-Agent": "alexxu-amd"}
    while True:
        try:
            response = requests.get(
                ROCM_TOOLKITS_URL, headers=headers, timeout=10
            )
            if response.status_code == 200:
                return _parse_rocm_toolkits(response.text.strip())
        except requests.RequestException:
            pass

        if time.monotonic() >= deadline:
            logger.warning(
                "Failed to fetch rocm_toolkits.txt after 10 minutes; "
                "using hardcoded fallback list."
            )
            return ROCM_TOOLKITS_FALLBACK

        time.sleep(5)


def _parse_rocm_toolkits(content: str) -> dict[str, str]:
    """Parse rocm_toolkits.txt (name: url lines) into a dict.

    Splits on the first colon only; the URL value (which also contains
    colons) is captured in full after that first separator.
    Example line: "ROCm Data Science: https://rocm.docs.amd.com/..."
    """
    result: dict[str, str] = {}
    for line in content.splitlines():
        line = line.strip()
        if not line:
            continue
        key, sep, value = line.partition(": ")
        if not sep:
            continue
        key = key.strip()
        value = value.strip()
        if key and value:
            result[key] = value
    return result if result else ROCM_TOOLKITS_FALLBACK


def _get_version_from_url(url: str) -> str:
    headers = {"User-Agent": "alexxu-amd"}
    try:
        retry_counter = 0
        response = requests.get(url, headers=headers)

        # Retry in case of failure
        while (response.status_code != 200) and (retry_counter <= MAX_RETRY):
            time.sleep(5)
            response = requests.get(url, headers=headers)

        if retry_counter > MAX_RETRY:
            raise requests.RequestException(
                "Unable to acquire version within MAX_RETRY!"
            )
        return response.text.strip()
    except requests.RequestException as e:
        print(f"Error in rocm-docs-core _get_version_from_url: {e}")
        return ""


def _parse_version(version_string: str) -> dict[str, str]:
    """Parse latest_version.txt and output a dictionary of site_name : latest_version.

    Example:
    {"ROCm": "7.0.2", "AI-Developer-Hub": "v7.0", "ROCm-DS": "25.05"}
    """
    header_latest_version_list = [
        site_version.split(":") for site_version in version_string.split("\n")
    ]
    return {
        site_version_pair[0].strip(): site_version_pair[1].strip()
        for site_version_pair in header_latest_version_list
    }


def _add_custom_context(
    app: Sphinx,  # noqa: ARG001
    pagename: str,  # noqa: ARG001
    templatename: str,  # noqa: ARG001
    context: dict[str, str | dict[str, str]],
    doctree: object,  # noqa: ARG001
) -> None:
    latest_version_list = _get_version_from_url(
        "https://raw.githubusercontent.com/ROCm/rocm-docs-core/new_data/latest_version.txt"
    )
    context["header_latest_version"] = _parse_version(latest_version_list)

    header_release_candidate_version = _get_version_from_url(
        "https://raw.githubusercontent.com/ROCm/rocm-docs-core/new_data/release_candidate.txt"
    )
    context["header_release_candidate_version"] = (
        header_release_candidate_version
    )

    google_site_verification_content = _get_version_from_url(
        "https://raw.githubusercontent.com/ROCm/rocm-docs-core/data/google_site_verification.txt"
    )
    context["google_site_verification_content"] = (
        google_site_verification_content
    )

    context["header_rocm_toolkits"] = _get_rocm_toolkits()


def _update_repo_opts(srcdir: str, theme_opts: dict[str, Any]) -> None:
    default_branch_options: dict[str, Any] = {
        "use_edit_page_button": False,
    }
    try:
        url, branch = util.get_branch(srcdir)
        default_branch_options.update(
            {
                "repository_url": url,
                "repository_branch": branch,
                "path_to_docs": util.get_path_to_docs(srcdir),
            }
        )
    except util.InvalidGitRepositoryError:
        logger.warning("Not in a Git Directory, disabling repository buttons")

    for key, val in default_branch_options.items():
        theme_opts.setdefault(key, val)


def _update_banner(
    flavor: str, version_type: util.VersionType, theme_opts: dict[str, Any]
) -> None:
    if flavor != "rocm":
        return

    if version_type == util.VersionType.OTHER_LATEST_RELEASE:
        return

    announcement_info: str
    preview_version = _get_version_from_url(
        "https://raw.githubusercontent.com/ROCm/rocm-docs-core/new_data/preview_version.txt"
    )
    latest_version_string = _get_version_from_url(
        "https://raw.githubusercontent.com/ROCm/rocm-docs-core/new_data/latest_version.txt"
    )
    latest_version = _parse_version(latest_version_string).get("rocm", "latest")
    if version_type == util.VersionType.RELEASE_CANDIDATE:
        announcement_info = "This page contains changes for a test release of ROCm. Read the <a id='rocm-banner' href='https://rocm.docs.amd.com/en/latest/'>latest Linux release of ROCm documentation</a> for your production environments."
    elif version_type == util.VersionType.OLD_RELEASE:
        announcement_info = "This is not the latest version of ROCm documentation. See <a id='rocm-banner' href='https://rocm.docs.amd.com/en/latest/'>ROCm documentation</a> for the latest version."
    elif version_type == util.VersionType.DEVELOPMENT:
        announcement_info = "This page contains proposed changes for a future release of ROCm. Read the <a id='rocm-banner' href='https://rocm.docs.amd.com/en/latest/'>latest Linux release of ROCm documentation</a> for your production environments."
    elif version_type == util.VersionType.PREVIEW:
        announcement_info = f"This is ROCm {preview_version} technology preview release documentation. For production use, refer to <a id='rocm-banner' href='https://rocm.docs.amd.com/en/latest/'>ROCm {latest_version} documentation</a>."
    elif version_type == util.VersionType.ROCM_LATEST_RELEASE:
        announcement_info = f"The ROCm {preview_version} technology preview release documentation is available at <a id='rocm-banner' href='https://rocm.docs.amd.com/en/{preview_version}-preview/'>ROCm Preview documentation</a>. For production use, continue to use ROCm {latest_version} documentation."

    theme_opts.setdefault("announcement", announcement_info)


def _update_theme_options(app: Sphinx) -> None:
    theme_opts = get_theme_options_dict(app)
    _update_repo_opts(str(app.srcdir), theme_opts)

    supported_flavors = [
        "rocm",
        "local",
        "instinct",
        "rocm-docs-home",
        "rocm-blogs",
        "generic",
        "rocm-ds",
        "ai-developer-hub",
        "rocm-ls",
        "gsplat",
        "rocm-rag",
        "amdgpu",
        "rocm-finance",
        "rocm-simulation",
        "rocm-llmext",
    ]
    flavor = theme_opts.get("flavor", "rocm")
    if flavor not in supported_flavors:
        logger.error(
            f'Unsupported theme flavor "{flavor}", must be one of: {supported_flavors}.\n'
            "Using flavor={supported_flavors[0]}"
        )
        flavor = supported_flavors[0]
        theme_opts["flavor"] = flavor

    # Set default generic theme options
    if flavor == "generic":
        theme_opts.setdefault("header_title", "")
        theme_opts.setdefault("header_link", "#")
        theme_opts.setdefault("version_list_link", None)
        theme_opts.setdefault("nav_secondary_items", {})
        theme_opts.setdefault("license_link", None)
        theme_opts.setdefault("license_text", "")

    theme_opts.setdefault(
        "article_header_start",
        ["components/toggle-primary-sidebar.html", "breadcrumbs.html"],
    )

    if hasattr(app.config, "projects_version_type"):
        _update_banner(flavor, app.config.projects_version_type, theme_opts)

    # Default the download, edit, and fullscreen buttons to off
    for button in ["download", "edit_page", "fullscreen"]:
        theme_opts.setdefault(f"use_{button}_button", False)

    if theme_opts.get("link_main_doc", True):
        theme_opts.setdefault("navbar_center", []).insert(
            0, "components/left-side-menu"
        )

    header_latest_version = _get_version_from_url(
        "https://raw.githubusercontent.com/ROCm/rocm-docs-core/new_data/latest_version.txt"
    )

    header_release_candidate_version = _get_version_from_url(
        "https://raw.githubusercontent.com/ROCm/rocm-docs-core/new_data/release_candidate.txt"
    )

    default_config_opts = {
        "html_show_sphinx": False,
        "html_favicon": "https://www.amd.com/content/dam/code/images/favicon/favicon.ico",
        "notfound_context": {"title": "404 - Page Not Found"},
        "notfound_template": "404.html",
        "html_context": {
            "header_latest_version": _parse_version(header_latest_version),
            "header_release_candidate_version": header_release_candidate_version,
            "header_rocm_toolkits": _get_rocm_toolkits(),
        },
    }
    for key, default in default_config_opts.items():
        if not config_provided_by_user(app, key):
            setattr(app.config, key, default)


def setup(app: Sphinx) -> dict[str, Any]:
    """Set up the module as a Sphinx extension."""
    app.add_js_file(
        "https://download.amd.com/js/analytics/analyticsinit.js",
        priority=999_999,
        loading_method="async",
    )
    app.add_js_file("code_word_breaks.js", loading_method="async")
    app.add_js_file("renameVersionLinks.js", loading_method="async")
    app.add_js_file("rdcMisc.js", loading_method="async")
    app.add_js_file("theme_mode_captions.js", loading_method="async")
    app.add_js_file("search.js", loading_method="defer")
    here = Path(__file__).parent.resolve()
    theme_path = here / "rocm_docs_theme"
    app.add_html_theme("rocm_docs_theme", str(theme_path))
    for css in [
        "custom.css",
        "rocm_header.css",
        "rocm_footer.css",
        "fonts.css",
    ]:
        app.add_css_file(css)

    app.connect("html-page-context", _add_custom_context)
    app.connect("builder-inited", _update_theme_options)

    # Add theme option declarations
    app.add_config_value("header_title", "", "html")
    app.add_config_value("header_link", "#", "html")
    app.add_config_value("version_list_link", None, "html")
    app.add_config_value("nav_secondary_items", {}, "html", [dict])
    app.add_config_value("license_link", None, "html")
    app.add_config_value("license_text", "", "html")

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Defines fixtures for interacting with projects.py"""

from __future__ import annotations

from collections.abc import Callable

import pytest

import rocm_docs.projects


@pytest.fixture
def mock_projects(
    monkeypatch: pytest.MonkeyPatch,
) -> Callable[[dict[str, rocm_docs.projects._Project]], None]:
    """Mock the list of projects loaded by projects.py"""

    def do_mock(projects: dict[str, rocm_docs.projects._Project]) -> None:
        monkeypatch.setattr(
            "rocm_docs.projects._load_projects",
            lambda *_, **__: projects,
        )

    return do_mock


@pytest.fixture
def mocked_projects(
    mock_projects: Callable[[dict[str, rocm_docs.projects._Project]], None],
) -> dict[str, rocm_docs.projects._Project]:
    """Standard mocked projects for tests"""
    projects = {
        "a": rocm_docs.projects._Project(
            "https://example.com/a", [], "", [], ".doxygen/docBin/html"
        ),
        "b": rocm_docs.projects._Project("https://example.com/b", [], "", []),
    }
    mock_projects(projects)
    return projects



"""Utilities for formatting text."""

from __future__ import annotations

from typing import Any

import re
from collections.abc import Generator, Iterable
from dataclasses import dataclass
from pathlib import Path


class Formatter:
    """Formatting class for string substitution and comments parsing."""

    @dataclass
    class _Replacement:
        loc: tuple[int, int]
        text: str

    def __init__(self, context: dict[str, Any]):
        """Initialize Formatter."""
        self.directive_pattern: re.Pattern[str] = re.compile(
            r"(?P<prefix>\$)?\{(?P<directive>[a-zA-z][a-zA-Z0-9_]+)"
            r"(?:\:(?P<parameter>[a-zA-Z0-9_\-\.]+))?\}"
        )
        self.context = context

    def _format_simple(
        self, directive: str, parameter: str | None, loc: tuple[int, int]
    ) -> _Replacement | None:
        # Cannot have a parameter
        if parameter is not None:
            return None
        return self._Replacement(loc, self.context[directive])

    def _format_project(
        self, _: str, parameter: str | None, loc: tuple[int, int]
    ) -> _Replacement | None:
        # Parameter is required
        if parameter is None:
            return None
        if parameter not in self.context["projects"]:
            return None
        return self._Replacement(loc, self.context["projects"][parameter])

    def _format_directive(self, match: re.Match[str]) -> _Replacement | None:
        # As a special case allow `{branch}` and `url` to alias `${branch}`
        # and '${url}' respectively for backwards compatibility.
        # Otherwise the '$' is required
        if match["prefix"] is None and match["directive"] not in [
            "branch",
            "url",
        ]:
            return None

        if match["directive"] in ["branch", "url"]:
            return self._format_simple(
                match["directive"], match["parameter"], match.span()
            )

        if match["directive"] == "project":
            return self._format_project(
                match["directive"], match["parameter"], match.span()
            )

        return None

    def _replacements(self, line: str) -> Generator[_Replacement, None, None]:
        for match in self.directive_pattern.finditer(line):
            replacement = self._format_directive(match)
            if replacement is not None:
                yield replacement

    def format_line(self, line: str) -> str:
        """Substitute variable references into line.

        References of the form ${<variable>} and ${directive:param}
        are substituted
        >>> f = Formatter(
        ...     {
        ...         "branch": "develop",
        ...         "url": "https://example.com",
        ...         "projects": {"project": "https://project.com"},
        ...     }
        ... )
        >>> f.format_line('my branch is ${branch}, {branch} also works')
        'my branch is develop, develop also works'
        >>> f.format_line('Url: ${url} or {url}')
        'Url: https://example.com or https://example.com'
        >>> f.format_line('- url: ${project:project}')
        '- url: https://project.com'

        Unknown references are not replaced.
        >>> f.format_line('{invalid}')
        '{invalid}'
        """
        result: str = ""
        end: int = 0
        for replacement in self._replacements(line):
            assert replacement.loc[0] >= end
            result += line[end : replacement.loc[0]] + replacement.text
            end = replacement.loc[1]

        result += line[end:]
        return result

    def skip_comments(self, lines: Iterable[str]) -> Generator[str, None, None]:
        """Returns a sequence that skips lines as long as they start with '#'.

        Lines after the first "non-comment" line are returned as-is.
        >>> f = Formatter({})
        >>> for l in f.skip_comments(
        ...     ["#comment 0", "# comment 1", "text", "# will not be skipped"]
        ... ):
        ...     print(l)
        text
        # will not be skipped
        """
        iterator = iter(lines)
        for line in iterator:
            if not line.startswith("#"):
                yield line
                break
        yield from iterator


def format_toc(
    input_path: Path, output_path: Path, context: dict[str, Any]
) -> None:
    """Format the input table of contents with additional information."""
    formatter = Formatter(context)
    with (
        open(input_path, encoding="utf-8") as toc_in,
        open(output_path, "w", encoding="utf-8") as toc_out,
    ):
        for line in formatter.skip_comments(toc_in):
            toc_out.write(formatter.format_line(line))



"""Defines Helpers for capturing and working with logs in tests"""

from __future__ import annotations

import types
from typing import Literal, NamedTuple

import contextlib
import copy
import itertools
import logging
import unittest.mock
from collections.abc import Callable, Iterable, Iterator

import pytest

TestPhase = Literal["setup", "call", "teardown"]


class LogHandlerStackToken(NamedTuple):
    when: TestPhase
    stacklevel: int


class LogStackFixture:
    def __init__(self, caplog: pytest.LogCaptureFixture):
        self._caplog: pytest.LogCaptureFixture = caplog
        self._log_stack: list[list[logging.LogRecord]] = []

    def push(self, *, when: TestPhase = "call") -> LogHandlerStackToken:
        # This should be a single atomic operation
        records = self._caplog.get_records(when)
        if len(records) > 0:
            self._log_stack.append(copy.copy(records))
            self._caplog.clear()
        return LogHandlerStackToken(when, len(self._log_stack))

    def pop(self, token: LogHandlerStackToken) -> Iterable[logging.LogRecord]:
        return itertools.chain(
            *self._log_stack[token.stacklevel :],
            self._caplog.get_records(token.when),
        )

    class Scope:
        def __init__(
            self,
            fixture: LogStackFixture,
            token: LogHandlerStackToken,
            callbacks: contextlib.ExitStack,
        ) -> None:
            self._fixture = fixture
            self._token = token
            self._callbacks = callbacks

        def add(
            self, callback: Callable[[Iterable[logging.LogRecord]], None]
        ) -> None:
            self._callbacks.callback(
                lambda: callback(self._fixture.pop(self._token))
            )

        def pop_all(self) -> LogStackFixture.Scope:
            return LogStackFixture.Scope(
                self._fixture, self._token, self._callbacks.pop_all()
            )

        def close(self) -> None:
            self._callbacks.close()

        def __enter__(self) -> LogStackFixture.Scope:
            return self

        def __exit__(
            self,
            type: type,
            value: BaseException,
            traceback: types.TracebackType,
        ) -> bool | None:
            return self._callbacks.__exit__(type, value, traceback)

    def new_scope(self, *, when: TestPhase = "call") -> Scope:
        return self.Scope(self, self.push(when=when), contextlib.ExitStack())


@pytest.fixture
def log_handler_stack(
    caplog: pytest.LogCaptureFixture, _with_sphinx_logs: None
) -> LogStackFixture:
    """A fixture allowing scoped log handling"""
    return LogStackFixture(caplog)


@pytest.fixture
def expected_logs_impl() -> set[logging.LogRecord]:
    """Implementation of expected_logs"""
    return set()


@pytest.fixture
def _no_unexpected_warnings_impl(
    request: pytest.FixtureRequest,
    log_handler_stack: LogStackFixture,
    expected_logs_impl: set[logging.LogRecord],
) -> Iterator[None]:
    if "_no_unexpected_warnings" not in request.fixturenames:
        yield
        return

    class Validator:
        def __init__(self) -> None:
            self.logrecord: tuple[str, str, str] | None = None

        def validate(self, records: Iterable[logging.LogRecord]) -> None:
            for r in records:
                if r.levelno < logging.WARNING:
                    continue
                if r not in expected_logs_impl:
                    self.logrecord = (
                        r.name,
                        r.levelname,
                        r.getMessage(),
                    )
                    return

    validator = Validator()
    with log_handler_stack.new_scope() as scope:
        scope.add(validator.validate)
        yield

    invert = request.node.get_closest_marker("meta_invert_fixture") is not None
    if validator.logrecord is None:
        if not invert:
            return

        pytest.fail("Did not find any unexpected logs")

    if not invert:
        pytest.fail(f"Unexpected log: {validator.logrecord}")


# Must request no_unexpected_warnings_ to ensure its teardown runs *after*
# other fixtures that request expected_logs
@pytest.fixture
def expected_logs(
    _no_unexpected_warnings_impl: None,
    expected_logs_impl: set[logging.LogRecord],
) -> set[logging.LogRecord]:
    """The set of expected logs, see no_unexpected_warnings"""
    return expected_logs_impl


@pytest.fixture
def _no_unexpected_warnings(
    _no_unexpected_warnings_impl: None,
) -> None:
    """Raise an error if an unexpected log is found

    all expected warnings should be added to `expected_logs`
    """
    return


class ExpectLogFixture:
    """Fixture class for expect_log"""

    def __init__(
        self,
        log_handler_stack: LogStackFixture,
        expected_logs: set[logging.LogRecord],
    ) -> None:
        """Initialize the fixture"""
        self._log_handler_stack = log_handler_stack
        self._expected_logs = expected_logs

    class Validator:
        def __init__(
            self,
            scope: LogStackFixture.Scope,
            expected: tuple[str, int, str],
            *,
            required: bool,
            capture_all: bool,
            expected_logs: set[logging.LogRecord],
        ) -> None:
            self._scope = scope
            self._expected = expected
            self._expected_logs = expected_logs
            self.required = required
            self.capture_all = capture_all
            self._found = False

            self._scope.add(self._validate_logs)

        @property
        def found(self) -> bool:
            return self._found

        def __enter__(self) -> ExpectLogFixture.Validator:
            self._scope.__enter__()
            return self

        def __exit__(
            self,
            type: type,
            value: BaseException,
            traceback: types.TracebackType,
        ) -> bool | None:
            return self._scope.__exit__(type, value, traceback)

        def _validate_logs(
            self,
            records: Iterable[logging.LogRecord],
        ) -> None:
            self._found = False
            for r in records:
                if self._expected != (r.name, r.levelno, r.getMessage()):
                    continue

                self._found = True
                self._expected_logs.add(r)
                if not self.capture_all:
                    return

            if not self.required or self._found:
                return

            lvl = logging.getLevelName(self._expected[1])
            pytest.fail(
                f"Expected log {(self._expected[0], lvl, self._expected[2])}, but it wasn't found"
            )

    def __call__(
        self,
        name: str,
        level: str,
        msg: str,
        *,
        required: bool = True,
        capture_all: bool = False,
        when: TestPhase = "call",
    ) -> ExpectLogFixture.Validator:
        """Register a new handler. See expect_log for more details"""
        levelno = logging.getLevelName(level)
        if not isinstance(levelno, int):
            raise ValueError(f'Unknown log level "{level}"')

        return self.Validator(
            self._log_handler_stack.new_scope(when=when),
            (name, levelno, msg),
            required=required,
            capture_all=capture_all,
            expected_logs=self._expected_logs,
        )


@pytest.fixture
def expect_log(
    log_handler_stack: LogStackFixture, expected_logs: set[logging.LogRecord]
) -> ExpectLogFixture:
    """Register a verifier for a log message

    Each call to expect_log(name, level, msg) will verify that such a log is
    emitted at least once. This log will not trigger unknown warning checks.
    Set capture_all to True to capture all warnings that match, not just the first.
    """
    return ExpectLogFixture(log_handler_stack, expected_logs)


@pytest.fixture
def _with_sphinx_logs(monkeypatch: pytest.MonkeyPatch) -> None:
    """Fixture to enable capturing sphinx logs in pytest"""
    monkeypatch.setattr("sphinx.util.logging.setup", unittest.mock.Mock())


__all__ = [
    "_no_unexpected_warnings",
    "_no_unexpected_warnings_impl",
    "_with_sphinx_logs",
    "expect_log",
    "expected_logs",
    "expected_logs_impl",
    "log_handler_stack",
]



extensions = ["rocm_docs", "rocm_docs.doxygen", "sphinxcontrib.doxylink"]
html_theme = "rocm_docs_theme"

external_projects_current_project = "a"

doxygen_project = {
    "name": "ROCm Docs Core Test Project - Extension",
    "path": "doxygen/xml",
}
doxysphinx_enabled = True
